/*++

Copyright (c) 2010-11 OKI SEMICONDUCTOR CO.,LTD.  All Rights Reserved

Module Name:

	gpio_sample.c

Abstract:

	Sample program for GPIO on Intel(R) Platform Controller Hub EG20T.

Environment:

	Win32 console multi-threaded application
	BigCove

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.

--*/

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strsafe.h>
#include <initguid.h>
#include <setupapi.h>

#include "ioh_gpio_common.h"
#include "ioh_gpio_ioctls.h"

extern BOOL RunGPIOSample(__in LPCTSTR DriverName, __in int argc, __in_ecount(argc) char *argv[]);
extern BOOL GetDevicePath(__in LPGUID pGuid, __out PCHAR pDeviceName);

int __cdecl
main(
	__in int argc,
	__in_ecount(argc) char *argv[]
	)
{
	CHAR DevicePath[MAX_PATH];

	if((argc == 2) || (argc == 3))
	{
		GetDevicePath((LPGUID)&GUID_DEVINTERFACE_IOHGPIO, DevicePath);
		RunGPIOSample(DevicePath, argc-1, argv+1);
	}
	else
	{
		printf("\n");
		printf("Usage: %s port [output]\n", argv[0]);
		printf(" port       Specify the port number (port=0-11 decimal).\n");
		printf(" output     Specify the port value (output=0 | 1).\n");
		printf("            When this is not specified, the port direction is set input mode.\n");
		printf("            When this is specified, the port is output mode.\n");
		printf("\n");
	}

	return 0;
}

// Get GPIO Device Path
BOOL GetDevicePath(LPGUID pGuid,PCHAR pDeviceName)
{
	HDEVINFO	hardwareDeviceInfo = NULL;
	SP_DEVICE_INTERFACE_DATA			deviceInterfaceData ={0};
	PSP_DEVICE_INTERFACE_DETAIL_DATA	deviceInterfaceDetailData = NULL;
	ULONG								predictedLength = 0;
	ULONG								requiredLength = 0,bytes=0;
	int i =0;
	BOOL bRet = FALSE;

		//
	// Open a handle to the device interface information set of all
	// present	system class interfaces.
	//

	hardwareDeviceInfo = SetupDiGetClassDevs(
						  pGuid,
						  NULL,
						  NULL,
						 (DIGCF_PRESENT | // Only Devices present
						 DIGCF_DEVICEINTERFACE)); // Function class devices.

	if( INVALID_HANDLE_VALUE == hardwareDeviceInfo )
	{
		printf("SetupDiGetClassDevs failed: %d\n", (INT)GetLastError());
		return bRet;
	}

	deviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

	//
	// Enumerate devices
	//

	do {

		if (SetupDiEnumDeviceInterfaces (hardwareDeviceInfo,
								 0, // No care about specific PDOs
								 pGuid,
								 i, //
								 &deviceInterfaceData))
			{

			//
			// Allocate a function class device data structure to
			// receive the information about this particular device.
			//

			//
			// First find out required length of the buffer
			//

			if(!SetupDiGetDeviceInterfaceDetail (
					hardwareDeviceInfo,
					&deviceInterfaceData,
					NULL, // probing so no output buffer yet
					0, // probing so output buffer length of zero
					&requiredLength,
					NULL))
			{ // not interested in the specific dev-node
				if(ERROR_INSUFFICIENT_BUFFER != GetLastError())
				{
					printf("SetupDiGetDeviceInterfaceDetail failed %d\n", (INT)GetLastError());
					SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
					break;
				}
			}

			predictedLength = requiredLength;

			deviceInterfaceDetailData = malloc (predictedLength);

			if(deviceInterfaceDetailData)
			{
				deviceInterfaceDetailData->cbSize =
								sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);
			} else
			{
				printf("Couldn't allocate %d bytes for device interface details.\n", (INT)predictedLength);
				SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
				break;
			}


			if (! SetupDiGetDeviceInterfaceDetail (
					   hardwareDeviceInfo,
					   &deviceInterfaceData,
					   deviceInterfaceDetailData,
					   predictedLength,
					   &requiredLength,
					   NULL))
			{
				printf("Error in SetupDiGetDeviceInterfaceDetail\n");
				SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
				free (deviceInterfaceDetailData);
				break;
			}

			printf("%d) %s\n", ++i,
					deviceInterfaceDetailData->DevicePath);

			strcpy_s(pDeviceName,MAX_PATH,deviceInterfaceDetailData->DevicePath);
			bRet = TRUE;
			break;
		}
		else if (ERROR_NO_MORE_ITEMS != GetLastError())
		{
			free (deviceInterfaceDetailData);
			deviceInterfaceDetailData = NULL;
			break;
		}
	} while (FALSE);

	SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);

	if(!deviceInterfaceDetailData)
	{
		printf("No device interfaces present\n");
		bRet = FALSE;
	}

	return bRet;
}

// Run GPIO Device Driver Check
BOOL RunGPIOSample(
	__in LPCTSTR DriverName,
	__in int argc,
	__in_ecount(argc) char *argv[]
)
{
	BOOL bRc = FALSE;
	HANDLE hDevice;
	DWORD portnum=0;
	DWORD portval=0;
	DWORD errNum = 0;

	ioh_gpio_reqt objGpioReqtOut;
	ioh_gpio_reqt objGpioReqtIn;

	// DeviceIoControl parameters
	LPVOID lpOutBuffer;
	DWORD nOutBufferSize;
	LPVOID lpInBuffer;
	DWORD nInBufferSize;
	DWORD dwBytesReturned;
	LPOVERLAPPED lpOverlapped;

	printf("\n=========================\n");
	printf("GPIO Device Driver Check(%s) \n", DriverName);
	printf("=========================\n");

	if (argc>0) {
		portnum = (UCHAR)atoi(argv[0]);
		printf("Specified port number : %d\n", (INT)portnum);
	}

	printf("\n*** CreateFile(%s) \n", DriverName);

	hDevice = CreateFile(DriverName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED,NULL);
	if(hDevice != INVALID_HANDLE_VALUE){
		printf("  Sucess CreateFile %s (hDevice:%08X)\n", DriverName, (UINT)hDevice);

		if (argc == 1) {
			printf("\n*** DeviceIoControl(code:IOCTL_GPIO_DIRECTION <0, GPIO%d, INPUT, 0>)\n", (INT)portnum);
			objGpioReqtOut.port=0;
			objGpioReqtOut.pins=(1<<portnum);
			objGpioReqtOut.mode=INPUT_SEL;
			objGpioReqtOut.enable=0;
			lpOutBuffer = (LPVOID)&objGpioReqtOut;
			nOutBufferSize = sizeof(objGpioReqtOut);
			lpInBuffer = (LPVOID)&objGpioReqtIn;
			nInBufferSize = sizeof(objGpioReqtIn);
			memset(lpInBuffer,0, nInBufferSize); // clear for in buffer
			dwBytesReturned = 0;
			lpOverlapped = NULL;

			bRc = DeviceIoControl(hDevice,IOCTL_GPIO_DIRECTION,lpOutBuffer,nOutBufferSize,lpInBuffer,nInBufferSize,&dwBytesReturned,lpOverlapped);
			if(bRc == TRUE) {
				printf("  Sucess DeviceIoControl\n");
				printf("      dwBytesReturned = %ld, output buffer = {port:%lx, pins:%lx, mode:%llx, enable:%lx}\n",
					dwBytesReturned, objGpioReqtIn.port, objGpioReqtIn.pins, objGpioReqtIn.mode, objGpioReqtIn.enable);

				printf("\n*** DeviceIoControl(code:IOCTL_GPIO_READ <0, GPIO%d, 0, 0>)\n", (INT)portnum);
				objGpioReqtOut.port=0;
				objGpioReqtOut.pins=(1<<portnum);
				objGpioReqtOut.mode=0;
				objGpioReqtOut.enable=0;
				lpOutBuffer = (LPVOID)&objGpioReqtOut;
				nOutBufferSize = sizeof(objGpioReqtOut);
				lpInBuffer = (LPVOID)&objGpioReqtIn;
				nInBufferSize = sizeof(objGpioReqtIn);
				memset(lpInBuffer,0, nInBufferSize); // clear for in buffer
				dwBytesReturned = 0;
				lpOverlapped = NULL;

				bRc = DeviceIoControl(hDevice,IOCTL_GPIO_READ,lpOutBuffer,nOutBufferSize,lpInBuffer,nInBufferSize,&dwBytesReturned,lpOverlapped);
				if(bRc == TRUE) {
					printf("  Sucess DeviceIoControl\n");
					printf("      dwBytesReturned = %ld, output buffer = {port:%lx, pins:%lx, mode:%llx, enable:%lx}\n",
					dwBytesReturned, objGpioReqtIn.port, objGpioReqtIn.pins, objGpioReqtIn.mode, objGpioReqtIn.enable);
					portval = (objGpioReqtIn.pins & (1<<portnum));
 					printf("\nGPIO%d = %s\n", (INT)portnum, (portval ? "HIGH" : "LOW"));
 				}
			}
		}
		else if (argc == 2) {
			portval = (UCHAR)atoi(argv[1]);
			printf("Specified port value : %s\n", (portval ? "HIGH" : "LOW"));

			printf("\n*** DeviceIoControl(code:IOCTL_GPIO_DIRECTION <0, GPIO%d, OUTPUT, 0>)\n", (INT)portnum);
			objGpioReqtOut.port=0;
			objGpioReqtOut.pins=(1<<portnum);
			objGpioReqtOut.mode=OUTPUT_SEL;
			objGpioReqtOut.enable=0;
			lpOutBuffer = (LPVOID)&objGpioReqtOut;
			nOutBufferSize = sizeof(objGpioReqtOut);
			lpInBuffer = (LPVOID)&objGpioReqtIn;
			nInBufferSize = sizeof(objGpioReqtIn);
			memset(lpInBuffer,0, nInBufferSize); // clear for in buffer
			dwBytesReturned = 0;
			lpOverlapped = NULL;

			bRc = DeviceIoControl(hDevice,IOCTL_GPIO_DIRECTION,lpOutBuffer,nOutBufferSize,lpInBuffer,nInBufferSize,&dwBytesReturned,lpOverlapped);
			if(bRc == TRUE) {
				printf("  Sucess DeviceIoControl\n");
				printf("      dwBytesReturned = %ld, output buffer = {port:%lx, pins:%lx, mode:%llx, enable:%lx}\n",
					dwBytesReturned, objGpioReqtIn.port, objGpioReqtIn.pins, objGpioReqtIn.mode, objGpioReqtIn.enable);

				printf("\n*** DeviceIoControl(code:IOCTL_GPIO_WRITE <0, GPIO%d, %s, 0>)\n", (INT)portnum, (portval ? "HIGH" : "LOW"));
				objGpioReqtOut.port=0;
				objGpioReqtOut.pins=(1<<portnum);
				objGpioReqtOut.mode=(portval ? GPIO_HIGH : GPIO_LOW);
				objGpioReqtOut.enable=0;
				lpOutBuffer = (LPVOID)&objGpioReqtOut;
				nOutBufferSize = sizeof(objGpioReqtOut);
				lpInBuffer = (LPVOID)&objGpioReqtIn;
				nInBufferSize = sizeof(objGpioReqtIn);
				memset(lpInBuffer,0, nInBufferSize); // clear for in buffer
				dwBytesReturned = 0;
				lpOverlapped = NULL;

				bRc = DeviceIoControl(hDevice,IOCTL_GPIO_WRITE,lpOutBuffer,nOutBufferSize,lpInBuffer,nInBufferSize,&dwBytesReturned,lpOverlapped);
				if(bRc == TRUE) {
					printf("  Sucess DeviceIoControl\n");
					printf("      dwBytesReturned = %ld, output buffer = {port:%lx, pins:%lx, mode:%llx, enable:%lx}\n",
						dwBytesReturned, objGpioReqtIn.port, objGpioReqtIn.pins, objGpioReqtIn.mode, objGpioReqtIn.enable);
				}
			}
		}

		if (bRc == FALSE) {
			errNum = GetLastError();
			printf("  Failded DeviceIoControl GetLastError:%d\n", (INT)errNum);
		}

		if(CloseHandle(hDevice)) {
			printf("\n*** CloseHandle(hDevice:%08X)\n", (UINT)hDevice);
			if(CloseHandle(hDevice)) {
				printf("  Sucess CloseHandle\n");
				bRc =  TRUE;
			}
		}
	}
	else {
		printf("Failue CreateFile %s (LastError:%d)\n", DriverName, (INT)GetLastError());
	}
	return bRc;
}

