/*++

Copyright (c) 2010-11 OKI SEMICONDUCTOR CO.,LTD.  All Rights Reserved

Module Name:

	i2c_sample.c

Abstract:

	Sample program for I2C on Intel(R) Platform Controller Hub EG20T.

Environment:

	Win32 console multi-threaded application


 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.

--*/

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strsafe.h>
#include <initguid.h>
#include <setupapi.h>

#include "ioh_i2c_common.h"
#include "ioh_i2c_ioctls.h"

BOOL RunI2CSample(__in LPCTSTR DriverName, __in int argc, __in_ecount(argc) char *argv[]);
BOOL GetDevicePath(__in LPGUID pGuid, __out PCHAR pDeviceName);

//I2C Defines
#define I2C_EEPROM_ADDR_MASK	(0x7FFF)	// The maxmum memory size is 32K bytes

#define EEPROM_ADDRESS 			(0)			// A1 and A0 = 00-03
#define EEPROM_ID				(0x50)		// Decided by the connected EEPROM.

#define DEVICE_ADDRESS			(EEPROM_ID | EEPROM_ADDRESS)
#define NORMAL_MODE 			(0x0)
#define MAX_BUFFER_SIZE 		(64)


int __cdecl
main(
	__in int argc,
	__in_ecount(argc) char *argv[]
	)
{
	CHAR DevicePath[MAX_PATH];

	if(argc >= 2)
	{
		GetDevicePath((LPGUID)&GUID_DEVINTERFACE_IOHI2C, DevicePath);
		RunI2CSample(DevicePath, argc-1, argv+1);
	}
	else
	{
		printf("\n");
		printf("Usage: %s address [data [data [data [...]]]]\n", argv[0]);
		printf(" address    Specify the address of the I2C EEPROM (address=0000-%04X).\n", I2C_EEPROM_ADDR_MASK);
		printf(" data       Specify the write data to the I2C EEPROM (data=00-FF).\n");
		printf("            When the data is not specified, it reads 8 bytes from I2C EEPROM.\n");
		printf("            It is possible to spcify the data up to 8 bytes or less.\n");
		printf("\n");
	}
	return 0;
}

// Get I2C Device Path
BOOL GetDevicePath(LPGUID pGuid,PCHAR pDeviceName)
{
	HDEVINFO	hardwareDeviceInfo = NULL;
	SP_DEVICE_INTERFACE_DATA			deviceInterfaceData ={0};
	PSP_DEVICE_INTERFACE_DETAIL_DATA	deviceInterfaceDetailData = NULL;
	ULONG								predictedLength = 0;
	ULONG								requiredLength = 0,bytes=0;
	int i =0;
	BOOL bRet = FALSE;

	//
	// Open a handle to the device interface information set of all
	// present	system class interfaces.
	//

	hardwareDeviceInfo = SetupDiGetClassDevs(
						  pGuid,
						  NULL,
						  NULL,
						 (DIGCF_PRESENT | // Only Devices present
						 DIGCF_DEVICEINTERFACE)); // Function class devices.

	if( INVALID_HANDLE_VALUE == hardwareDeviceInfo )
	{
		printf("SetupDiGetClassDevs failed: %d\n", (INT)GetLastError());
		return bRet;
	}

	deviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

	//
	// Enumerate devices
	//

	do {

		if (SetupDiEnumDeviceInterfaces (hardwareDeviceInfo,
								 0, // No care about specific PDOs
								 pGuid,
								 i, //
								 &deviceInterfaceData))
			{

			//
			// Allocate a function class device data structure to
			// receive the information about this particular device.
			//

			//
			// First find out required length of the buffer
			//

			if(!SetupDiGetDeviceInterfaceDetail (
					hardwareDeviceInfo,
					&deviceInterfaceData,
					NULL, // probing so no output buffer yet
					0, // probing so output buffer length of zero
					&requiredLength,
					NULL))
			{ // not interested in the specific dev-node
				if(ERROR_INSUFFICIENT_BUFFER != GetLastError())
				{
					printf("SetupDiGetDeviceInterfaceDetail failed %d\n", (INT)GetLastError());
					SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
					break;
				}
			}

			predictedLength = requiredLength;

			deviceInterfaceDetailData = malloc (predictedLength);

			if(deviceInterfaceDetailData)
			{
				deviceInterfaceDetailData->cbSize =
								sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);
			}
			else {
				printf("Couldn't allocate %d bytes for device interface details.\n", (INT)predictedLength);
				SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
				break;
			}


			if (! SetupDiGetDeviceInterfaceDetail (
					   hardwareDeviceInfo,
					   &deviceInterfaceData,
					   deviceInterfaceDetailData,
					   predictedLength,
					   &requiredLength,
					   NULL))
			{
				printf("Error in SetupDiGetDeviceInterfaceDetail\n");
				SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
				free (deviceInterfaceDetailData);
				break;
			}

			printf("%d) %s\n", ++i,
					deviceInterfaceDetailData->DevicePath);

			strcpy_s(pDeviceName,MAX_PATH,deviceInterfaceDetailData->DevicePath);
			bRet = TRUE;
			break;
		}
		else if (ERROR_NO_MORE_ITEMS != GetLastError())
		{
			free (deviceInterfaceDetailData);
			deviceInterfaceDetailData = NULL;
			break;
		}
	} while (FALSE);

	SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);

	if(!deviceInterfaceDetailData)
	{
		printf("No device interfaces present\n");
		bRet = FALSE;
	}

	return bRet;
}


// Run I2C Device Driver Check
BOOL RunI2CSample(
	__in LPCTSTR DriverName,
	__in int argc,
	__in_ecount(argc) char *argv[]
)
{
	UCHAR msgbuf[1];
	HANDLE hDevice;
	BOOL bRc = FALSE;
	DWORD errNum = 0;
	UINT temp=0;
	USHORT address = 0;
	UCHAR value = 0;
	INT mbytes = 8;
	INT i;

	// DeviceIoControl parameters
	DWORD dwIoControlCode;
	LPVOID lpInBuffer;
	DWORD nInBufferSize;
	LPVOID lpOutBuffer;
	DWORD nOutBufferSize;
	DWORD dwBytesReturned;
	LPOVERLAPPED lpOverlapped;

	UCHAR InBuffer[MAX_BUFFER_SIZE];
	UCHAR OutBuffer[MAX_BUFFER_SIZE];

	// for I2C Device Driver API
	IOH_I2C_MSG i2cConfig = {0};

	printf("\n=========================\n");
	printf("I2C Device Driver Check(%s) \n", DriverName);
	printf("=========================\n");

	if (argc>0) {
		sscanf_s(argv[0], "%x", &temp);
		address = (USHORT)temp & I2C_EEPROM_ADDR_MASK;
		printf("Specified address : %04X\n", address);
	}

	printf("\n*** CreateFile(%s) \n", DriverName);
	hDevice = CreateFile(DriverName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0,NULL);
	if(hDevice != INVALID_HANDLE_VALUE){
		printf("  Sucess CreateFile %s (handle:%08X)\n", DriverName, (UINT)hDevice);

		printf("\n*** DeviceIoControl(code:IOCTL_I2C_CONFIG <SlaveAddress:%02Xh, Flags:0>)\n", DEVICE_ADDRESS);
		msgbuf[0] = NORMAL_MODE;
		i2cConfig.SlaveAddress = DEVICE_ADDRESS;
		i2cConfig.MsgData = msgbuf;
		i2cConfig.MsgLength = 1;
		i2cConfig.Flags = 0;

		lpOutBuffer = (LPVOID)&i2cConfig;
		nOutBufferSize = sizeof(i2cConfig);
		dwBytesReturned = 0;
		lpOverlapped = NULL;
		bRc = DeviceIoControl(hDevice,IOCTL_I2C_CONFIG,lpOutBuffer,nOutBufferSize,NULL,0,&dwBytesReturned,lpOverlapped);
		if(bRc == TRUE) {
				printf("  Sucess DeviceIoControl\n");
				printf("      dwBytesReturned = %ld\n", dwBytesReturned);

			if (argc>1) {
				printf("\n*** I2C EEPROM WRITING(code:IOCTL_I2C_WRITE <%02Xh> <%02Xh> ...)\n", (UINT)((address >> 8) & 0xFF), (UINT)(address & 0xFF));
				OutBuffer[0] = (UCHAR)((address >> 8) & 0xFF);	// 1st Address
				OutBuffer[1] = (UCHAR)(address & 0xFF);			// 2nd Address

				mbytes = min((INT)argc - 1, 8);
				for(i=0; i<mbytes; i++) {
					sscanf_s(argv[i+1], "%x", &temp);
					value = (UCHAR)temp;
					printf("Specified value : %02X\n", value);
					OutBuffer[2+i] = value;	// Data
				}

				nOutBufferSize = 2 + mbytes;
				dwBytesReturned = 0;
				lpOverlapped = NULL;
				bRc = DeviceIoControl(hDevice,IOCTL_I2C_WRITE,OutBuffer,nOutBufferSize,NULL,0,&dwBytesReturned,lpOverlapped);
				if(bRc == TRUE) {
					printf("  Sucess DeviceIoControl\n");
					printf("      dwBytesReturned = %ld\n", dwBytesReturned);

					Sleep(5); // Wait for memory write cycle. Refer to spce of EE2PROM
				}
			}
			if(bRc == TRUE) {
				printf("\n*** I2C EEPROM READING(code:IOCTL_I2C_WRITE <%02Xh> <%02Xh>)\n", (UINT)((address >> 8) & 0xFF), (UINT)(address & 0xFF));
				OutBuffer[0] = (UCHAR)((address >> 8) & 0xFF);	// 1st Address
				OutBuffer[1] = (UCHAR)(address & 0xFF);			// 2nd Address
				nOutBufferSize = 2;
				dwBytesReturned = 0;
				lpOverlapped = NULL;
				bRc = DeviceIoControl(hDevice,IOCTL_I2C_WRITE,OutBuffer,nOutBufferSize,NULL,0,&dwBytesReturned,lpOverlapped);
				if(bRc == TRUE) {
					printf("  Sucess DeviceIoControl\n");
					printf("      dwBytesReturned = %ld\n", dwBytesReturned);

					printf("\n*** I2C EEPROM READING(code:IOCTL_I2C_READ %d byte)\n", mbytes);
					memset(InBuffer, 0, sizeof(InBuffer));
					nInBufferSize = mbytes;
					dwBytesReturned = 0;
					lpOverlapped = NULL;
					bRc = DeviceIoControl(hDevice,IOCTL_I2C_READ,NULL,0,InBuffer,nInBufferSize,&dwBytesReturned,lpOverlapped);
					if(bRc == TRUE) {
						printf("  Sucess DeviceIoControl\n");
						printf("      dwBytesReturned = %ld\n", dwBytesReturned);
						for(i=0; i<mbytes; i++) {
							printf("      InBuffer[%d] = %02X\n", i,(UINT)InBuffer[i]);
						}
					}
				}
			}
		}

		if (bRc == FALSE) {
			errNum = GetLastError();
			printf("  Failded DeviceIoControl GetLastError:%d\n", (INT)errNum);
		}

		printf("\n*** CloseHandle(handle:%08X)\n", (UINT)hDevice);
		if(CloseHandle(hDevice)) {
			printf("  Sucess CloseHandle\n");
			bRc =  TRUE;
		}
		else {
			errNum = GetLastError();
			printf("  Failded CloseHandle GetLastError:%d\n", (INT)errNum);
		}
	}
	else {
		printf("Failue CreateFile %s (LastError:%d)\n", DriverName, (INT)GetLastError());
	}

	return bRc;
}


