/*++

Copyright (c) 2010-11 OKI SEMICONDUCTOR CO.,LTD.  All Rights Reserved

Module Name:

	spi_sample.c

Abstract:

	Sample program for SPI on Intel(R) Platform Controller Hub EG20T.

Environment:

	Win32 console multi-threaded application

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.

--*/

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strsafe.h>
#include <initguid.h>
#include <setupapi.h>

#include "ioh_spi_common.h"
#include "ioh_spi_ioctls.h"

BOOL RunSPISample(__in LPCTSTR DriverName, __in int argc, __in_ecount(argc) char *argv[]);
BOOL GetDevicePath(__in LPGUID pGuid, __out PCHAR pDeviceName);

//SPI Defines
#define SPI_EEPROM_ADDR_MASK		(0x1FF)

#define SPI_WRITE_INSTRUCTION_8BIT	(0x2)
#define SPI_READ_INSTRUCTION_8BIT	(0x3)
#define SPI_WRITE_ENABLE_8BIT		(0x6)

#define MAX_BUFFER_SIZE (64)


int __cdecl
main(
	__in int argc,
	__in_ecount(argc) char *argv[]
	)
{
	CHAR DevicePath[MAX_PATH];

	if(argc >= 2)
	{
		GetDevicePath((LPGUID)&GUID_DEVINTERFACE_IOHSPI, DevicePath);
		RunSPISample(DevicePath, argc-1, argv+1);
	}
	else
	{
		printf("\n");
		printf("Usage: %s address [data [data [data [...]]]]\n", argv[0]);
		printf(" address    Specify the address of the SPI EEPROM (address=000-%03X).\n", SPI_EEPROM_ADDR_MASK);
		printf(" data       Specify the write data to the SPI EEPROM (data=00-FF).\n");
		printf("            When the data is not specified, it reads 8 bytes from SPI EEPROM.\n");
		printf("            It is possible to spcify the data up to 8 bytes or less.\n");
		printf("\n");
	}
	return 0;
}

// Get SPI Device Path
BOOL GetDevicePath(LPGUID pGuid,PCHAR pDeviceName)
{
	HDEVINFO	hardwareDeviceInfo = NULL;
	SP_DEVICE_INTERFACE_DATA			deviceInterfaceData ={0};
	PSP_DEVICE_INTERFACE_DETAIL_DATA	deviceInterfaceDetailData = NULL;
	ULONG								predictedLength = 0;
	ULONG								requiredLength = 0,bytes=0;
	int i =0;
	BOOL bRet = FALSE;

	//
	// Open a handle to the device interface information set of all
	// present	system class interfaces.
	//

	hardwareDeviceInfo = SetupDiGetClassDevs(
						  pGuid,
						  NULL,
						  NULL,
						 (DIGCF_PRESENT | // Only Devices present
						 DIGCF_DEVICEINTERFACE)); // Function class devices.

	if( INVALID_HANDLE_VALUE == hardwareDeviceInfo )
	{
		printf("SetupDiGetClassDevs failed: %d\n", (INT)GetLastError());
		return bRet;
	}

	deviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

	//
	// Enumerate devices
	//

	do {

		if (SetupDiEnumDeviceInterfaces (hardwareDeviceInfo,
								 0, // No care about specific PDOs
								 pGuid,
								 i, //
								 &deviceInterfaceData))
			{

			//
			// Allocate a function class device data structure to
			// receive the information about this particular device.
			//

			//
			// First find out required length of the buffer
			//

			if(!SetupDiGetDeviceInterfaceDetail (
					hardwareDeviceInfo,
					&deviceInterfaceData,
					NULL, // probing so no output buffer yet
					0, // probing so output buffer length of zero
					&requiredLength,
					NULL))
			{ // not interested in the specific dev-node
				if(ERROR_INSUFFICIENT_BUFFER != GetLastError())
				{
					printf("SetupDiGetDeviceInterfaceDetail failed %d\n", (INT)GetLastError());
					SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
					break;
				}
			}

			predictedLength = requiredLength;

			deviceInterfaceDetailData = malloc (predictedLength);

			if(deviceInterfaceDetailData)
			{
				deviceInterfaceDetailData->cbSize =
								sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);
			}
			else {
				printf("Couldn't allocate %d bytes for device interface details.\n", (INT)predictedLength);
				SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
				break;
			}


			if (! SetupDiGetDeviceInterfaceDetail (
					   hardwareDeviceInfo,
					   &deviceInterfaceData,
					   deviceInterfaceDetailData,
					   predictedLength,
					   &requiredLength,
					   NULL))
			{
				printf("Error in SetupDiGetDeviceInterfaceDetail\n");
				SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
				free (deviceInterfaceDetailData);
				break;
			}

			printf("%d) %s\n", ++i,
					deviceInterfaceDetailData->DevicePath);

			strcpy_s(pDeviceName,MAX_PATH,deviceInterfaceDetailData->DevicePath);
			bRet = TRUE;
			break;
		}
		else if (ERROR_NO_MORE_ITEMS != GetLastError())
		{
			free (deviceInterfaceDetailData);
			deviceInterfaceDetailData = NULL;
			break;
		}
	} while (FALSE);

	SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);

	if(!deviceInterfaceDetailData)
	{
		printf("No device interfaces present\n");
		bRet = FALSE;
	}

	return bRet;
}


// Run SPI Device Driver Check
BOOL RunSPISample(
	__in LPCTSTR DriverName,
	__in int argc,
	__in_ecount(argc) char *argv[]
)
{
	UCHAR msgbuf[1];
	HANDLE hDevice;
	BOOL bRc = FALSE;
	DWORD errNum = 0;
	UINT temp=0;
	USHORT address = 0;
	UCHAR address_low = 0;
	UCHAR address_a8 = 0;
	UCHAR value = 0;
	INT mbytes = 8;
	INT i;

	// DeviceIoControl parameters
	DWORD dwIoControlCode;
	LPVOID lpInBuffer;
	DWORD nInBufferSize;
	LPVOID lpOutBuffer;
	DWORD nOutBufferSize;
	DWORD dwBytesReturned;

	UCHAR InBuffer[MAX_BUFFER_SIZE];
	UCHAR OutBuffer[MAX_BUFFER_SIZE];

	// for SPI Device Driver API
	IOH_SPI_CONFIG spiConfig = {0};
	IOH_SPI_RW ioData = {0};

	printf("\n=========================\n");
	printf("SPI Device Driver Check(%s) \n", DriverName);
	printf("=========================\n");

	if (argc>0) {
		sscanf_s(argv[0], "%x", &temp);
		address = (USHORT)temp & SPI_EEPROM_ADDR_MASK;
		printf("Specified address : %04X\n", address);
		address_low = (UCHAR)(address & 0xFF);
		address_a8 = (UCHAR)((address>>8) & 0x1);
	}

	printf("\n*** CreateFile(%s) \n", DriverName);
	hDevice = CreateFile(DriverName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0,NULL);
	if(hDevice != INVALID_HANDLE_VALUE) {
		printf("  Sucess CreateFile %s (hDevice:%08X)\n", DriverName, (UINT)hDevice);

		printf("\n*** DeviceIoControl(code:IOCTL_SIP_CONFIG <1, 25000, IOH_SPI_8_BPW, IOH_SPI_MODE_0>)\n", DriverName);
		spiConfig.SlaveNo	  = 1;
		spiConfig.BaudRate	  = 25000;
		spiConfig.BitsPerWord = IOH_SPI_8_BPW;
		spiConfig.Mode		  = IOH_SPI_MODE_0;

		dwIoControlCode = IOCTL_SPI_CONFIG;
		lpOutBuffer = (LPVOID)&spiConfig;
		nOutBufferSize = sizeof(spiConfig);
		dwBytesReturned = 0;
		bRc = DeviceIoControl(hDevice,dwIoControlCode,lpOutBuffer,nOutBufferSize,NULL,0,&dwBytesReturned,NULL);
		if(bRc == TRUE) {
			printf("  Sucess DeviceIoControl\n");
			printf("      dwBytesReturned = %ld\n", dwBytesReturned);

			if (argc>1) {
				printf("\n*** SPI EEPROM WRITING(code:IOCTL_SPI_WRITE <WRITE_ENABLE(%02Xh)>)\n", SPI_WRITE_ENABLE_8BIT);
				OutBuffer[0] = SPI_WRITE_ENABLE_8BIT;
				ioData.SpiConfig = spiConfig;
				ioData.Buffer = OutBuffer;
				ioData.nSize = 1;
				dwBytesReturned = 0;
				bRc = DeviceIoControl(hDevice,IOCTL_SPI_WRITE,&ioData,sizeof(IOH_SPI_RW),NULL,0,&dwBytesReturned,NULL);
				if(bRc == TRUE) {
					printf("  Sucess DeviceIoControl\n");
					printf("      dwBytesReturned = %ld\n", dwBytesReturned);

					printf("\n*** SPI EEPROM WRITING(code:IOCTL_SPI_WRITE <WRITE_INSTRUCTION((%02Xh)> <LOW ADDRESS(%02Xh)> ...)\n",
						(UINT)((UCHAR)SPI_WRITE_INSTRUCTION_8BIT | (address_a8 << 3)), (UINT)address_low);
					OutBuffer[0] = (UCHAR)SPI_WRITE_INSTRUCTION_8BIT | (address_a8 << 3);
					OutBuffer[1] = address_low;

					mbytes = min((INT)argc - 1, 8);
					for(i=0; i<mbytes; i++) {
						sscanf_s(argv[i+1], "%x", &temp);
						value = (UCHAR)temp;
						printf("Specified value : %02X\n", value);
						OutBuffer[2+i] = value;	// Data
					}

					ioData.SpiConfig = spiConfig;
					ioData.Buffer = OutBuffer;
					ioData.nSize = 2 + mbytes;
					dwBytesReturned = 0;
					bRc = DeviceIoControl(hDevice,IOCTL_SPI_WRITE,&ioData,sizeof(IOH_SPI_RW),NULL,0,&dwBytesReturned,NULL);
					if(bRc == TRUE) {
						printf("  Sucess DeviceIoControl\n");
						printf("      dwBytesReturned = %ld\n", dwBytesReturned);

						Sleep(500);
					}
				}
			}
			if(bRc == TRUE) {
				printf("\n*** SPI EEPROM READING(code:IOCTL_SPI_READ <READ_INSTRUCTION(%02Xh)> <LOW ADDRESS(%02Xh)>; Read %d byte)\n",
					(UINT)((UCHAR)SPI_READ_INSTRUCTION_8BIT | (address_a8 << 3)), (UINT)address_low, mbytes);
				OutBuffer[0] = (UCHAR)SPI_READ_INSTRUCTION_8BIT | (address_a8 << 3);
				OutBuffer[1] = address_low;
				ioData.SpiConfig = spiConfig;
				ioData.Buffer = OutBuffer;
				ioData.nSize = 2+mbytes;
				nInBufferSize = sizeof(InBuffer);
				dwBytesReturned = 0;
				bRc = DeviceIoControl(hDevice,IOCTL_SPI_READ,&ioData,sizeof(IOH_SPI_RW),InBuffer,nInBufferSize,&dwBytesReturned,NULL);
				if(bRc == TRUE) {
					printf("  Sucess DeviceIoControl\n");
					printf("      dwBytesReturned = %ld\n", dwBytesReturned);

					/* The 1st and 2nd byte data in the InBuffer[] are invalid. */
					for (i=2; i<mbytes+2; i++) {
						printf("      InBuffer[%d] = %02X\n", i, (UINT)InBuffer[i]);
					}
				}
			}
		}

		if (bRc == FALSE) {
			errNum = GetLastError();
			printf("  Failded DeviceIoControl GetLastError:%d\n", (INT)errNum);
		}

		printf("\n*** CloseHandle(hDevice:%08X)\n", (UINT)hDevice);
		if(CloseHandle(hDevice)) {
			printf("  Sucess CloseHandle\n");
			bRc =  TRUE;
		}
		else {
			errNum = GetLastError();
			printf("  Failded CloseHandle GetLastError:%d\n", (INT)errNum);
		}
	}
	else {
		printf("Failue CreateFile %s (LastError:%d)\n", DriverName, (INT)GetLastError());
	}

	return bRc;
}


